from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from rest_framework import serializers as rf

import aeat

from . import complex_types, fields


def make_aeat_request(service_name, data):
    '''
    Helper to make AEAT requests

    :rtype: aduanet.aeat.Result
    '''

    cert_path = settings.AEAT_CERT_PATH
    key_path = settings.AEAT_KEY_PATH

    if not cert_path:
        raise ImproperlyConfigured('AEAT_CERT_PATH required')

    if not key_path:
        raise ImproperlyConfigured('AEAT_KEY_PATH required')

    config = aeat.Config(service_name, test_mode=settings.AEAT_TEST_MODE)
    ctrl = aeat.Controller.build_from_config(config, cert_path, key_path)
    return ctrl.request(data)


class AEATRequest(rf.Serializer):
    service_name = None

    def save(self):
        return make_aeat_request(self.service_name, self.data)

    def to_representation(self, instance):
        instance = super().to_representation(instance)

        if settings.AEAT_TEST_MODE:
            instance['TesIndMES18'] = '0'

        return instance


class ENSQuerySerializer(AEATRequest):
    service_name = 'ens_query'

    TraModAtBorHEA76 = fields.RequiredStr(help_text='Transport mode at border. EG 1')
    ExpDatOfArr = fields.RequiredStr(help_text='Estimated date of arrival. EG 20110809')
    ConRefNum = fields.RequiredStr(help_text='Transport identifier. EG 9294408')


class ENSForkSerializer(AEATRequest):
    service_name = 'ens_fork'


class MessageMixin(rf.Serializer):
    '''Common attributes'''
    MesSenMES3 = fields.NotRequiredStr(max_length=35, read_only=True,
                                       default=settings.AEAT_VAT_NUMBER,
                                       help_text='Message Sender (VAT Number). EG. 89890001K')
    MesRecMES6 = fields.NotRequiredStr(max_length=35, default='NICA.ES', read_only=True,
                                       help_text='EG NICA.ES (default)')
    DatOfPreMES9 = fields.AEATDateField(required=True, allow_null=False,
                                        help_text='Date of preparation. EG 101010 (YYMMDD)')
    TimOfPreMES10 = fields.AEATTimeField(required=True, allow_null=False,
                                         help_text='Time of preparation. EG 1010 (HHMM)')
    MesIdeMES19 = fields.RequiredStr(max_length=14, help_text='Message identification. '
                                                              'EG 09ES112222110 (like Id)')

    TRACONCO1 = complex_types.TraderConsignor(required=True)
    TRACONCE1 = complex_types.TraderConsignee(required=True)
    GOOITEGDS = complex_types.GoodsItem(required=True, many=True)
    ITI = complex_types.Itinerary(required=False, many=True)
    TRAREP = complex_types.TraderRepresentative(required=False)
    PERLODSUMDEC = complex_types.PersonLodgingSummaryDeclaration(required=False)
    SEAID529 = complex_types.SealsIdentity(required=False, many=True)
    CUSOFFFENT730 = complex_types.CustomsOfficeFirstEntry(required=True)
    CUSOFFSENT740 = complex_types.CustomsOfficeSubsequentEntry(many=True)
    TRACARENT601 = complex_types.TraderEntryCarrier(required=False)


class ENSPresentationSerializer(MessageMixin, AEATRequest):
    service_name = 'ens_presentation'

    MesTypMES20 = fields.NotRequiredStr(default='CC315A', read_only=True,
                                        help_text='Message type. EG CC315A')
    HEAHEA = complex_types.ENSPresentationHeader(required=True)


class ENSModificationSerializer(MessageMixin, AEATRequest):
    service_name = 'ens_modification'
    NOTPAR670 = complex_types.NotifyParty(required=True)
    MesTypMES20 = fields.NotRequiredStr(default='CC313A', read_only=True,
                                        help_text='Message type. EG CC313A')
    HEAHEA = complex_types.ENSModificationHeader(required=True)


class EXSSerializer(MessageMixin, AEATRequest):
    service_name = 'exs_presentation'

    Id = fields.RequiredStr(max_length=14, source='MesIdeMES19',
                            help_text='Message identification. (like Id)')
    NifDeclarante = fields.NotRequiredStr(max_length=14, read_only=True,
                                          default=settings.AEAT_VAT_NUMBER)
    NombreDeclarante = fields.NotRequiredStr(max_length=14, read_only=True,
                                             default=settings.AEAT_LEGAL_NAME)
    MesTypMES20 = fields.NotRequiredStr(default='CC615A', read_only=True,
                                        help_text='Message type')
    HEAHEA = complex_types.EXSHeader(required=True)



import os
from unittest.mock import Mock

import pytest

from lxml import etree
from zeep import Client


@pytest.fixture
def admin_request(admin_user):
    def func(method, url=None, payload=None):
        from rest_framework.test import APIRequestFactory, force_authenticate

        factory = APIRequestFactory()
        method_func = getattr(factory, method)
        request = method_func(url, payload, format='json')
        force_authenticate(request, user=admin_user)
        return request

    return func


@pytest.fixture
def certificate_example(resource_path):
    cert_path = resource_path(['certificate', 'cert.pem'])
    key_path = resource_path(['certificate', 'key.pem'])
    return cert_path, key_path


@pytest.fixture
def certificate_real():
    cert_path = os.environ.get('AEAT_CERT_PATH')
    key_path = os.environ.get('AEAT_KEY_PATH')

    return cert_path, key_path


@pytest.fixture
def resource_path():
    tests_path = os.path.dirname(os.path.abspath(__file__))
    resources_path = os.path.join(tests_path, 'resources')

    def get_path(args):
        return os.path.join(*[resources_path] + args)

    return get_path


@pytest.fixture
def request_etree_element():
    tests_path = os.path.dirname(os.path.abspath(__file__))
    resources_path = os.path.join(tests_path, 'resources', 'xml', 'requests')

    def get_xml(filename):
        path = os.path.join(*[resources_path] + [filename])
        with open(path, 'r') as xml_template:
            return etree.fromstring(xml_template.read())

    return get_xml


@pytest.fixture
def zeep_response(resource_path):
    def response_maker(wsdl, response, operation):
        wsdl_path = resource_path(['xml', 'xades', 'wsdl', wsdl])

        client = Client(wsdl_path, strict=False)

        operation = client.service._binding._operations[operation]
        response_path = resource_path(['xml', 'response', response])

        with open(response_path, 'rb') as f:
            aeat_response = Mock(status_code=200, headers={}, content=f.read())

            return client.service._binding.process_reply(
                client, operation, aeat_response)

    return response_maker



import logging

from . import zeep_plugins
from .wsdl import ADUANET_SERVICES
from requests import Session
from zeep import exceptions as zeep_exceptions
from zeep import Client, Transport, xsd


logger = logging.getLogger(__name__)


class Config:
    def __init__(self, service_name, test_mode):
        self.payload = ADUANET_SERVICES[service_name]
        self.test_mode = test_mode

    def __getattr__(self, name):
        if name == 'port':
            test, prod = self.payload['port_test'], self.payload['port_production']
            return test if self.test_mode else prod

        return self.payload[name]

    def __str__(self):
        return f'''
        Config: {self.verbose_name}
        WSDL: {self.wsdl}
        Operation: {self.operation}
        Port: {self.port}
        Service: {self.service}
        '''


class Result:
    def __init__(self, data, error, raw_request=None, raw_response=None):
        self.data = data
        self.error = error
        self.raw_request = raw_request
        self.raw_response = raw_response

    @property
    def valid(self):
        return self.error is None


def raw_response_parser(data):
    return Result(data, None)


def mrn_response_parser(data):
    '''
    Returns the MRN from the response or error response
    '''
    for item in data:
        mrn_el = item.find('DocNumHEA5')
        mrn = mrn_el.text if mrn_el is not None else None

        if mrn:
            return Result(mrn, None)

    return Result(None, 'AEAT response error')


DEFAULT_RESPONSE_PARSER = raw_response_parser

RESPONSE_PARSERS = {
    'IE315V4': mrn_response_parser,
    'IE313V4': mrn_response_parser,
}


class Controller:
    def __init__(self, client, config, raw_xml_plugin=None):
        self.client = client
        self.config = config
        self.raw_xml_plugin = raw_xml_plugin
        logger.info('Controller initialized with config %s', config)

    @classmethod
    def build_from_config(cls, config, pub_cert, priv_cert):
        '''
        Builds AEAT Controller with Plugins

        :param config: Preconfigured Config object
        :param pub_cert: Public certificate file path
        :param priv_cert: Private certificate file path

        :rtype: Controller
        '''
        session = Session()
        session.cert = (pub_cert, priv_cert)
        transport = Transport(session=session, operation_timeout=60)

        raw_xml_plugin = zeep_plugins.RawXMLPlugin()

        if config.signed:
            sign_plugin = zeep_plugins.SignMessage(pub_cert, priv_cert)
            plugins = [sign_plugin, raw_xml_plugin]
        else:
            plugins = [raw_xml_plugin]

        client = Client(config.wsdl, service_name=config.service, port_name=config.port,
                        transport=transport, strict=False, plugins=plugins)

        return cls(client, config, raw_xml_plugin=raw_xml_plugin)

    @property
    def operation(self):
        return getattr(self.client.service, self.config.operation)

    def get_response_parser(self):
        return RESPONSE_PARSERS.get(self.config.operation, DEFAULT_RESPONSE_PARSER)

    def request(self, payload):
        if self.config.signed:
            # Skip WSDL validation. Is added later by zeep_plugins.SignMessage
            payload['Signature'] = xsd.SkipValue

        try:
            data = self.operation(**payload)
        except zeep_exceptions.Fault as e:
            logger.info('AEAT request failed.', exc_info=True)
            result = Result(None, e.message)
        except zeep_exceptions.Error as e:
            logger.info('AEAT request failed.', exc_info=True)
            result = Result(None, 'Wrong AEAT response')
        except Exception as e:
            logger.critical('Unexpected exception', exc_info=True)
            result = Result(None, 'Unknown error')
        else:
            parser = self.get_response_parser()
            result = parser(data)

        if self.raw_xml_plugin:
            result.raw_request = self.raw_xml_plugin.last_sent
            result.raw_response = self.raw_xml_plugin.last_received

        return result



import os
from unittest.mock import Mock, patch

import pytest

import django

# Setup Django as soon as possible
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'django_test_app.settings')
django.setup()


from django.core.exceptions import ImproperlyConfigured  # isort:skip # NOQA
from django.test.utils import override_settings  # isort:skip # NOQA

from rest_framework import serializers as rf  # isort:skip # NOQA

from aeat import Result  # isort:skip # NOQA
from aeat.rest_framework import serializers  # isort:skip # NOQA


@patch('aeat.rest_framework.serializers.aeat.Config', Mock())
@patch('aeat.rest_framework.serializers.aeat.Controller')
def test_make_aeat_request(ctrl_builder):
    ctrl_mock = Mock()
    ctrl_builder.build_from_config.return_value = ctrl_mock
    serializers.make_aeat_request('ens_query', {'count': '2'})
    assert ctrl_mock.request.called_with({'count': '2'})


@override_settings(AEAT_CERT_PATH=None)
@patch('aeat.rest_framework.serializers.aeat.Config', Mock())
@patch('aeat.rest_framework.serializers.aeat.Controller')
def test_make_aeat_request_without_cert_path_env_var_fails(ctrl_builder):
    with pytest.raises(ImproperlyConfigured):
        serializers.make_aeat_request('test_service', data={})


@override_settings(AEAT_KEY_PATH=None)
@patch('aeat.rest_framework.serializers.aeat.Config', Mock())
@patch('aeat.rest_framework.serializers.aeat.Controller')
def test_make_aeat_request_without_key_path_env_var_fails(ctrl_builder):
    with pytest.raises(ImproperlyConfigured):
        serializers.make_aeat_request('test_service', data={})


@patch('aeat.rest_framework.serializers.make_aeat_request',
       lambda x, y: Result(data=None, error='NO REENVIABLE'))
def test_aeat_request_serializer_returns_result_failed_when_request_fails():
    serializer = serializers.AEATRequest()
    result = serializer.save()
    assert not result.valid


@patch('aeat.rest_framework.serializers.make_aeat_request',
       lambda x, y: Result(data='xyz', error=None, raw_response='xml...'))
def test_aeat_request_serializer_returns_result_if_request_is_successfull():
    serializer = serializers.AEATRequest()
    result = serializer.save()
    assert result.valid


def test_adds_test_indicator_when_test_mode_is_enabled():
    class TestSerializer(serializers.AEATRequest, rf.Serializer):
        pass

    serializer = TestSerializer(data={})
    serializer.is_valid(raise_exception=True)

    assert '0' == serializer.data['TesIndMES18']


@override_settings(AEAT_TEST_MODE=False)
def test_does_not_adds_test_indicator_in_production():
    class TestSerializer(serializers.AEATRequest, rf.Serializer):
        pass

    serializer = TestSerializer(data={})
    serializer.is_valid(raise_exception=True)

    assert 'TesIndMES18' not in serializer.data



from rest_framework import serializers as rf

from .fields import AEATDateTimeField, NotRequiredStr, RequiredStr


class ENSPresentationHeader(rf.Serializer):
    '''HEAHEAType'''
    RefNumHEA4 = RequiredStr(max_length=22, help_text='Reference Number. EG LRN000000041')
    TraModAtBorHEA76 = RequiredStr(max_length=2, help_text='Transport mode at border. EG 5.')
    IdeOfMeaOfTraCroHEA85 = NotRequiredStr(
        max_length=27, help_text='Identity of means of transport crossing border. EG 1111111')
    IdeOfMeaOfTraCroHEA85LNG = NotRequiredStr(help_text='Identity of means of transport '
                                                        'crossing border LNG. EG ES')
    TotNumOfIteHEA305 = rf.IntegerField(required=True, min_value=0, max_value=5,
                                        help_text='Total number of items. EG: 3')
    TotNumOfPacHEA306 = NotRequiredStr(max_length=7,
                                       help_text='Total number of packages. EG: 50')
    TotGroMasHEA307 = NotRequiredStr(help_text='Total gross mass. EG 10')
    DecPlaHEA394 = RequiredStr(help_text='Declaration place. EG Madrid')
    DecPlaHEA394LNG = RequiredStr(help_text='Declaration place LNG. EG ES')
    SpeCirIndHEA1 = NotRequiredStr(help_text='Specific Circumstance Indicator. EG A')
    TraChaMetOfPayHEA1 = NotRequiredStr(help_text='Transport charges / Method of Payment. EG C')
    ComRefNumHEA = NotRequiredStr(max_length=70, help_text='Commercial Reference Numer. EG a828rt')
    ConRefNumHEA = NotRequiredStr(max_length=35, help_text='Conveyance reference number. EG 7777b')
    PlaLoaGOOITE334 = NotRequiredStr(max_length=35, help_text='Place of loading. EG ESMadrid')
    PlaLoaGOOITE334LNG = NotRequiredStr(help_text='Place of loading LNG. EG ES')
    PlaUnlGOOITE334 = NotRequiredStr(max_length=35, help_text='Place of unloading. EG ESSegovia')
    CodPlUnHEA357LNG = NotRequiredStr(help_text='Place of unloading LNG. EG ES')
    DecDatTimHEA114 = AEATDateTimeField(
        required=True, help_text='Declaration date and time. EG 201207041455')


class ENSModificationHeader(ENSPresentationHeader):
    '''HEAHEAType with mandatory MRNType'''
    DocNumHEA5 = RequiredStr(help_text='Document/reference number')


class EXSHeader(rf.Serializer):
    '''HEAHEAType EXS'''
    RefNumHEA4 = RequiredStr(max_length=22, help_text='Reference Number. EG LRN000000041')
    CusSubPlaHEA66 = RequiredStr(max_length=22, help_text='Customs sub place. EG 4611ZZZ999')
    TotNumOfIteHEA305 = rf.IntegerField(required=True, min_value=0, max_value=5,
                                        help_text='Total number of items. EG: 3')
    TotNumOfPacHEA306 = NotRequiredStr(max_length=7,
                                       help_text='Total number of packages. EG: 50')
    TotGroMasHEA307 = NotRequiredStr(help_text='Total gross mass. EG 10')
    SpeCirIndHEA1 = NotRequiredStr(help_text='Specific Circumstance Indicator. EG A')
    DecDatTimHEA114 = AEATDateTimeField(
        required=True, help_text='Declaration date and time. EG 201207041455')


class TraderConsignor(rf.Serializer):
    '''TRACONCO1Type'''
    NamCO17 = NotRequiredStr(help_text='Name. EG JUAN CARLOS')
    StrAndNumCO122 = NotRequiredStr(help_text='Street and number. EG Almansa')
    PosCodCO123 = NotRequiredStr(help_text='Postal code. EG 28007')
    CitCO124 = NotRequiredStr(help_text='City. EG Madrid')
    CouCO125 = NotRequiredStr(help_text='Country code. EG ES')
    NADLNGCO = NotRequiredStr(help_text='NAD LNG. EG ES')
    TINCO159 = NotRequiredStr(help_text='TIN (Trader identification number). EG ESA08005688')


class TraderConsignee(rf.Serializer):
    '''TRACONCE1Type'''
    NamCE17 = NotRequiredStr(help_text='Name. EG luis')
    StrAndNumCE122 = NotRequiredStr(help_text='Street and number. EG cruz')
    PosCodCE123 = NotRequiredStr(help_text='Postal code. EG 28005')
    CitCE124 = NotRequiredStr(help_text='City. EG Madrid')
    CouCE125 = NotRequiredStr(help_text='Country code. EG ES')
    NADLNGCE = NotRequiredStr(help_text='NAD LNG. EG ES')
    TINCE159 = NotRequiredStr(help_text='TIN (Trader identification number). EG ESA08005688')


class ProducedDocumentsCertificates(rf.Serializer):
    '''PRODOCDC2Type'''
    DocTypDC21 = RequiredStr(max_length=4, help_text='Document type. EG Y022')
    DocRefDC23 = RequiredStr(max_length=35, help_text='Document reference. EG ESAEOC1')
    DocRefDCLNG = NotRequiredStr(help_text='Document reference LNG')


class SpecialMentions(rf.Serializer):
    '''SPEMENMT2Type'''
    AddInfCodMT23 = RequiredStr(max_length=5, help_text='Additional information coded')


class CommodityCode(rf.Serializer):
    '''COMCODGODITMType'''
    ComNomCMD1 = RequiredStr(min_length=4, max_length=8,
                             help_text='Combined Nomenclature. EG 123456')


class Container(rf.Serializer):
    '''CONNR2Type'''
    ConNumNR21 = RequiredStr(max_length=17, help_text='Container number')


class Package(rf.Serializer):
    '''PACGS2Type'''
    KinOfPacGS23 = RequiredStr(max_length=2, help_text='Kind of packages')
    NumOfPacGS24 = NotRequiredStr(max_length=5, help_text='Number of packages')
    NumOfPieGS25 = NotRequiredStr(max_length=5, help_text='Number of pieces')
    MarNumOfPacGSL21 = NotRequiredStr(max_length=140,
                                      help_text='Marks & numbers of packages (long)')
    MarNumOfPacGSL21LNG = RequiredStr(help_text='Marks & numbers of packages (long) LNG')


class GoodsItem(rf.Serializer):
    '''GOOITEGDSType'''
    IteNumGDS7 = rf.IntegerField(required=True, help_text='Item Number. EG 1')
    GooDesGDS23 = NotRequiredStr(help_text='Goods description. EG DESCRIPCION  PARTIDA1')
    GooDesGDS23LNG = NotRequiredStr(help_text='Goods description LNG. EG ES')
    GroMasGDS46 = NotRequiredStr(help_text='Gross mass. EG 100')
    MetOfPayGDI12 = NotRequiredStr(help_text='Transport charges / Method of payment')
    ComRefNumGIM1 = NotRequiredStr(help_text='Commercial Reference Number. EG REFERENCIACOMER1')
    UNDanGooCodGDI1 = NotRequiredStr(help_text='UN dangerous goods code (Numeric 4). Min=0')
    PlaLoaGOOITE333 = NotRequiredStr(help_text='Place of loading. Max 35. Min=0')
    PlaLoaGOOITE333LNG = NotRequiredStr(help_text='Place of loading LNG. Min=0')
    PlaUnlGOOITE333 = NotRequiredStr(help_text='Place of unloading. Max 35. Min=0')
    PlaUnlGOOITE333LNG = NotRequiredStr(help_text='Place of unloading LNG. Min=0')

    PRODOCDC2 = ProducedDocumentsCertificates(required=False, many=True)
    SPEMENMT2 = SpecialMentions(required=False)
    # TRACONCO2 - not in aeat example. It may not be needed
    COMCODGODITM = CommodityCode(required=False)
    # TRACONCE2 - not in aeat example. It may not be needed
    CONNR2 = Container(many=True, required=False)
    # IDEMEATRAGI970 - not in aeat example. It may not be needed
    PACGS2 = Package(many=True, required=False)
    # PRTNOT640 - not in aeat example. It may not be needed


class Itinerary(rf.Serializer):
    '''ITIType'''
    CouOfRouCodITI1 = RequiredStr(help_text='Country of routing code')


class PersonLodgingSummaryDeclaration(rf.Serializer):
    '''PERLODSUMDECType'''
    NamPLD1 = NotRequiredStr(help_text='Name')
    StrAndNumPLD1 = NotRequiredStr(help_text='Street and number')
    PosCodPLD1 = NotRequiredStr(help_text='Postal code')
    CitPLD1 = NotRequiredStr(help_text='City')
    CouCodPLD1 = NotRequiredStr(help_text='Country code')
    PERLODSUMDECLNG = NotRequiredStr(help_text='Language code')
    TINPLD1 = RequiredStr(min_length=3, max_length=17,
                          help_text='Trader indentification number')


class SealsIdentity(rf.Serializer):
    '''SEAID529Type'''
    SeaIdSEAID530 = RequiredStr(max_length=20, help_text='Seals identity')
    SeaIdSEAID530LNG = NotRequiredStr(help_text='Seals identity LNG')


class CustomsOfficeFirstEntry(rf.Serializer):
    '''CUSOFFFENT730Type'''
    RefNumCUSOFFFENT731 = RequiredStr(max_length=8, help_text='Reference number')
    ExpDatOfArrFIRENT733 = AEATDateTimeField(
        required=True, help_text='Expected date and time of arrival')


class CustomsOfficeSubsequentEntry(rf.Serializer):
    '''CUSOFFSENT740Type'''
    RefNumSUBENR909 = RequiredStr(max_length=8, help_text='Reference number')


class TraderEntryCarrier(rf.Serializer):
    '''TRACARENT601Type'''
    NamTRACARENT604 = NotRequiredStr(help_text='Name')
    StrNumTRACARENT607 = NotRequiredStr(help_text='Street and number')
    PstCodTRACARENT606 = NotRequiredStr(help_text='Postal code')
    CtyTRACARENT603 = NotRequiredStr(help_text='City')
    CouCodTRACARENT605 = NotRequiredStr(help_text='Country code')
    TRACARENT601LNG = NotRequiredStr(help_text='Language code')
    TINTRACARENT602 = NotRequiredStr(help_text='Trader identification number')


class TraderRepresentative(rf.Serializer):
    '''TRAREPType'''
    NamTRE1 = NotRequiredStr(help_text='Name')
    StrAndNumTRE1 = NotRequiredStr(help_text='Street and number')
    PosCodTRE1 = NotRequiredStr(help_text='Postal code')
    CitTRE1 = NotRequiredStr(help_text='City')
    CouCodTRE1 = NotRequiredStr(help_text='Country code')
    TRAREPLNG = NotRequiredStr(help_text='NAD LNG')
    TINTRE1 = NotRequiredStr(help_text='Trader indentification number')


class NotifyParty(rf.Serializer):
    '''NOTPAR670Type'''
    NamNOTPAR672 = NotRequiredStr(help_text='Name')
    StrNumNOTPAR673 = NotRequiredStr(help_text='Street and number')
    PosCodNOTPAR676 = NotRequiredStr(help_text='Postal code')
    CitNOTPAR674 = NotRequiredStr(help_text='City')
    CouCodNOTPAR675 = NotRequiredStr(help_text='Country code')
    NOTPAR670LNG = NotRequiredStr(help_text='NAD LNG')
    TINNOTPAR671 = NotRequiredStr(help_text='Trader indentification number')



import os

import pytest

import factories

import aeat


'''
Test tagged as functional are are not run by default (see setup.cfg)
Copy .env-sample to .env and set the certificate paths

Enable the env vars with ```export `cat .env````
then run pytest specifiying the tag ```pytest -m functional```

Be aware that they are run against the AEAT (in test mode) with your REAL certificate.
'''


@pytest.fixture
def make_aeat_test_controller(certificate_real):
    def make_controller_for_service(service_name):
        config = aeat.Config(service_name, test_mode=True)
        return aeat.Controller.build_from_config(config, *certificate_real)

    return make_controller_for_service


@pytest.mark.functional
def test_ens_presentation(make_aeat_test_controller):
    ctrl = make_aeat_test_controller('ens_presentation')
    result = ctrl.request(factories.ENSPresentationFactory(
        MesSenMES3=os.environ.get('AEAT_VAT_NUMBER', 'X12345678')
    ))
    assert result.valid, result.error

    assert 'OK' == result.data  # WIP. Not sure what the response is yet


@pytest.mark.functional
def test_ens_modification(make_aeat_test_controller):
    ctrl = make_aeat_test_controller('ens_modification')
    result = ctrl.request(factories.ENSPresentationFactory(
        MesSenMES3=os.environ.get('AEAT_VAT_NUMBER', 'X12345678')
    ))
    assert result.valid, result.error

    assert 'OK' == result.data  # WIP. Not sure what the response is yet


@pytest.mark.functional
def test_ens_query(make_aeat_test_controller):
    ctrl = make_aeat_test_controller('ens_query')

    result = ctrl.request(factories.ENSQueryFactory())

    assert result.valid
    assert result.data['TraModAtBorHEA76'] == '1'
    assert result.data.ExpDatOfArr == '20110809'
    assert result.data['ConRefNum'] == '9294408'
    assert 770 == len(result.data.IMPOPE)



from unittest.mock import Mock, PropertyMock, patch

import pytest

import factories
from zeep import exceptions as zeep_exceptions
from zeep import xsd

from aeat import Config, Controller, wsdl


def test_config_as_str():
    config = Config('ens_presentation', test_mode=True)
    assert 'Servicio de Presentación ENS V4.0' in config.__str__()


@pytest.mark.parametrize('signed', [True, False])
@patch('aeat.Client')
def test_controller_is_built_from_config_obj(client, signed):
    config = Mock(signed=signed)
    ctrl = Controller.build_from_config(config, Mock(), Mock())
    assert isinstance(ctrl, Controller)
    assert signed is ctrl.config.signed


@pytest.mark.parametrize('test_mode,expected_port', [
    (True, 'IE315V4Pruebas'),
    (False, 'IE315V4'),
])
def test_config_is_built_from_service_name(test_mode, expected_port):
    config = Config('ens_presentation', test_mode=test_mode)
    assert config.wsdl.endswith('IE315V4.wsdl')
    assert 'IE315V4' == config.operation
    assert 'IE315V4Service' == config.service
    assert expected_port == config.port


@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_ens_query_with_valid_response(operation_patch, zeep_response):
    response = zeep_response(
        'wsdl_ens_query_ConsENSV3.wsdl', 'ens_query_valid.xml', 'ConsENSV3'
    )

    operation_patch.return_value = lambda **kwargs: response
    ctrl = Controller(Mock(), Mock())
    result = ctrl.request(factories.ENSQueryFactory())

    assert result.valid
    assert 2 == len(result.data['IMPOPE'])


@patch('aeat.Controller.operation')
def test_controller_marks_signature_as_skip_if_config_is_signed(operation_patch):
    config = Mock(signed=True)
    ctrl = Controller(Mock(), config)
    ctrl.request({'arg': 'x'})

    operation_patch.assert_called_with(arg='x', Signature=xsd.SkipValue)


@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_with_99999_error(operation_patch, zeep_response):
    def response():
        return zeep_response('wsdl_ens_presentation_IE315V4.wsdl',
                             'ens_presentation_error_99999.xml', 'IE315V4')

    operation_patch.return_value = lambda **kwargs: response()
    ctrl = Controller(Mock(), Mock())
    result = ctrl.request(factories.ENSQueryFactory())

    assert not result.valid
    assert 'Mensaje REENVIABLE. Codigo[99999].' == result.error
    assert result.data is None


@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_with_html_error(operation_patch, zeep_response):
    def response():
        return zeep_response('wsdl_ens_query_ConsENSV3.wsdl', 'unknown_certificate.html',
                             'ConsENSV3')

    operation_patch.return_value = lambda **kwargs: response()
    ctrl = Controller(Mock(), Mock())
    result = ctrl.request(factories.ENSQueryFactory())

    assert not result.valid
    assert 'Wrong AEAT response' == result.error
    assert result.data is None


def test_controller_operation():
    service = Mock()
    service.myoperation = Mock()
    ctrl = Controller(Mock(service=service), Mock(operation='myoperation'))
    assert service.myoperation == ctrl.operation


@pytest.mark.parametrize('service_name', [
    name for name, _ in wsdl.ADUANET_SERVICES.items()
])
def test_aduanet_services_configuration(service_name):
    config = Config(service_name, True)
    assert isinstance(config, Config)
    assert config.verbose_name is not None
    assert config.wsdl is not None
    assert config.operation is not None
    assert config.port is not None
    assert config.service is not None
    assert config.signed is not None
    assert config.port.endswith('Pruebas')


@pytest.mark.parametrize('detail,exception_cls', [
    ('Wrong AEAT response', zeep_exceptions.XMLSyntaxError),
    ('Wrong AEAT response', zeep_exceptions.ValidationError),
    ('Unknown error', Exception),
])
@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_operation_request_exception_handling(operation_patch, detail, exception_cls):
    def operation(arg, Signature):
        raise exception_cls

    operation_patch.return_value = operation

    config = Mock(signed=True)
    ctrl = Controller(Mock(), config)

    result = ctrl.request({'arg': 'x'})

    assert not result.valid
    assert detail == result.error


@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_with_ens_presentation_success_message(operation_patch, zeep_response):
    def response():
        return zeep_response('wsdl_ens_presentation_IE315V4.wsdl',
                             'ens_presentation_success_IE328V4Sal.xml', 'IE315V4')

    operation_patch.return_value = lambda **kwargs: response()
    ctrl = Controller(Mock(), Mock(operation='IE315V4'))
    result = ctrl.request(factories.ENSPresentationFactory())

    assert result.valid
    assert '17ES004311Z0000010' == result.data


@pytest.mark.parametrize('response_xml', [
    'ens_presentation_error_IE316V4Sal.xml',
    'ens_presentation_error_IE917V4Sal.xml',
])
@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_with_incorrect_responses(operation_patch, zeep_response, response_xml):
    def response():
        return zeep_response('wsdl_ens_presentation_IE315V4.wsdl', response_xml, 'IE315V4')

    operation_patch.return_value = lambda **kwargs: response()
    ctrl = Controller(Mock(), Mock(operation='IE315V4'))
    result = ctrl.request(factories.ENSPresentationFactory())

    assert not result.valid
    assert result.data is None
    assert 'AEAT response error' == result.error


@patch('aeat.Controller.operation', new_callable=PropertyMock)
def test_controller_result_includes_raw_request_and_response(operation_patch, zeep_response):
    def response():
        return zeep_response('wsdl_ens_presentation_IE315V4.wsdl',
                             'ens_presentation_success_IE328V4Sal.xml', 'IE315V4')

    operation_patch.return_value = lambda **kwargs: response()
    history_plugin = Mock(last_sent='xyz', last_received='zyx')
    ctrl = Controller(Mock(), Mock(operation='IE315V4'), history_plugin)
    result = ctrl.request(factories.ENSPresentationFactory())

    assert 'xyz' == result.raw_request
    assert 'zyx' == result.raw_response



BASE_ENS = 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/'

ADUANET_SERVICES = {
    'ens_fork': {
        'verbose_name': 'Servicio Solicitud Desvío ENS V4.0',
        'wsdl': f'{BASE_ENS}enswsv4/IE323V4.wsdl',
        'operation': 'IE323V4',
        'port_production': 'IE323V4',
        'port_test': 'IE323V4Pruebas',
        'service': 'IE323V4Service',
        'signed': True,
    },
    'ens_modification': {
        'verbose_name': 'Servicio Modificación ENS V4.0',
        'wsdl': f'{BASE_ENS}enswsv4/IE313V4.wsdl',
        'operation': 'IE313V4',
        'port_production': 'IE313V4',
        'port_test': 'IE313V4Pruebas',
        'service': 'IE313V4Service',
        'signed': True,
    },
    'ens_presentation': {
        'verbose_name': 'Servicio de Presentación ENS V4.0',
        'wsdl': f'{BASE_ENS}enswsv4/IE315V4.wsdl',
        'operation': 'IE315V4',
        'port_production': 'IE315V4',
        'port_test': 'IE315V4Pruebas',
        'service': 'IE315V4Service',
        'signed': True,
    },
    'ens_query': {
        'verbose_name': "Servicio Consulta de ENS's",
        'wsdl': f'{BASE_ENS}ensws/ConsENSV3.wsdl',
        'operation': 'ConsENSV3',
        'port_production': 'ConsENSV3',
        'port_test': 'ConsENSV3Pruebas',
        'service': 'ConsENSV3Service',
        'signed': False,
    },
    'exs_presentation': {
        'verbose_name': "Presentación de Declaración Sumaria de Salida",
        'wsdl': 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/adrx/ws/IE615V2.wsdl',  # NOQA
        'operation': 'IE615V2',
        'port_production': 'IE615V2',
        'port_test': 'IE615V2Pruebas',
        'service': 'IE615V2Service',
        'signed': True,
    },
}



import logging

from zeep import Plugin

from aeat import utils, xml_signing

logger = logging.getLogger(__name__)


class SignMessage(Plugin):
    def __init__(self, cert_path, key_path):
        self.cert_path = cert_path
        self.key_path = key_path

    def egress(self, envelope, http_headers, operation, binding_options):
        args = envelope, self.cert_path, self.key_path
        xml_signing.sign(*args)
        return envelope, http_headers


class RawXMLPlugin(object):
    '''
    Stores last request and response as str
    '''
    def __init__(self):
        self.last_sent = None
        self.last_received = None

    def ingress(self, envelope, http_headers, operation):
        self.last_received = utils.lxml_to_string(envelope)
        return envelope, http_headers

    def egress(self, envelope, http_headers, operation, binding_options):
        self.last_sent = utils.lxml_to_string(envelope)
        return envelope, http_headers



import os

'''Django test settings'''

INSTALLED_APPS = ['django_test_app']

SECRET_KEY = 'testing.'

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
PROJECT_DIR = os.path.dirname(BASE_DIR)

AEAT_CERT_PATH = os.path.join(PROJECT_DIR, 'aduanet', 'tests', 'resources', 'certificate', 'cert.pem')  # NOQA
AEAT_KEY_PATH = os.path.join(PROJECT_DIR, 'aduanet', 'tests', 'resources', 'certificate', 'key.pem')  # NOQA
AEAT_VAT_NUMBER = 'X12345678'
AEAT_LEGAL_NAME = 'Test Legal Name'
AEAT_TEST_MODE = True
